// -------------------------------------------------------------------------
/*
    bh_distort.vert.c

  Copyright (c) 2009-2014  Thomas Mueller

    This file is part of DssBH.

    DssBH is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DssBH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DssBH.  If not, see <http://www.gnu.org/licenses/>.
*/

/*!   \file   bh_distort.vert.c
      \brief  Vertex shader for blackhole distortion.
*/
// -------------------------------------------------------------------------

#version 330


#define    PI    3.14159265
#define invPI    0.31830989
#define invTwoPI 0.15915494
#define twoPI    6.28318531
#define PI_2     1.57079633

#define edlog210  0.301029996
#define edlg10    0.434294482

#define MAX_PHI  12.56637061
#define invMAX_PHI  0.07957747152

uniform mat4 mvp;

uniform sampler2D  ksiArray;
uniform sampler2D  ksiInsideArray;
uniform   float def_max_pointsize;
uniform   vec2   wSize;


uniform   float obsDist;  //scaled distance x_obs = rs/r_obs
uniform   float x0;
uniform   float a;
uniform   float b;
uniform   float beta;
uniform   float gamma;
uniform   int   order;

uniform   float ksiCrit;
uniform   mat4  rotmat;

layout(location = 0) in vec3 in_position;
layout(location = 1) in float absMag;
layout(location = 2) in float starTemp;

out float appMag;
out float logTempObs;
out float logTempStar;
out float one_over_mu;
out float chi;
out float kappa;


// -----------------------------------------------
//        transCartSphere
// -----------------------------------------------
void transCartSphere( const vec3 p, inout float theta, inout float phi )
{
    theta = atan(sqrt(p.x*p.x+p.y*p.y),p.z);
    phi   = atan(p.y,p.x);
}


// -----------------------------------------------
//        aberration
// -----------------------------------------------
float aberration ( float ksi )
{
    return  acos( (cos(ksi)+beta)/(1.0+beta*cos(ksi)) );
}

// -----------------------------------------------
//        calcNewRaDec
// -----------------------------------------------
bool calcNewRaDec ( in vec3 dir, out float newRa, out float newDe, out float phi, out float newKsi )
{
    vec3 ex,ey,ez;
    ex = vec3(1.0,0.0,0.0);
    ez = normalize(cross(ex,dir));
    ey = normalize(cross(ez,ex));

    phi = acos(dot(dir,ex));
    if (order==1)    phi = twoPI-phi;

    float k, h, w, x, y;
    newRa = newDe = 0.0;

    // freely falling
    if (obsDist>=1.0)
    {
        kappa = texture2D(ksiInsideArray, vec2(1.0/obsDist,phi*invMAX_PHI) ).r;
        w = sqrt(a*a+kappa*kappa-b*b);
        x = a*b+kappa*w;
        y = a*w-b*kappa;
        newKsi = atan(y,x);
        chi = texture2D(ksiInsideArray, vec2(1.0/obsDist,phi*invMAX_PHI) ).g;
    }
    else
    {
        newKsi = texture2D(ksiArray, vec2(obsDist,phi*invMAX_PHI) ).r;
        newKsi = aberration(newKsi);
        chi = texture2D(ksiArray, vec2(obsDist,phi*invMAX_PHI) ).g;
    }

    if (newKsi<ksiCrit)    return false;

    vec3  r = cos(newKsi)*ex + sin(newKsi)*ey;
    if (order==1)
        r = cos(newKsi)*ex - sin(newKsi)*ey;

    transCartSphere(r,newDe,newRa);
    return true;
}

// -----------------------------------------------
//        shiftTemp
// -----------------------------------------------
float zfactor( float cosksi )
{
    return sqrt(1.0-x0) - cosksi*sqrt(obsDist-x0);
}

// -----------------------------------------------
//        main
// -----------------------------------------------
void main()
{
    vec4 vert = vec4(0,0,0,1);
    float psc = 1000.0/in_position.x;
    float ra  = in_position.y;
    float de  = in_position.z;

    //vert.xyz = vec3( (1.0-ra*invTwoPI)*wSize.x, (de*invPI+0.5)*wSize.y, 0.01*(log(psc)) );

    vec4 dir = vec4(cos(de)*cos(ra),cos(de)*sin(ra),sin(de),0.0);
    dir = rotmat*dir;
    float newRa,newDe,phi,newKsi;
    bool ok = calcNewRaDec(dir.xyz,newRa,newDe,phi,newKsi);
    vert.xyz = vec3( (0.5+newRa*invTwoPI)*wSize.x, (1.0-newDe*invPI)*wSize.y, 0.01*log(psc) );

    // Apparent magnitude of the star.
    appMag = absMag + 10.0 - 5.0*log(psc)*edlg10;

    if (appMag<14.0 && ok)
        gl_PointSize = def_max_pointsize;
    else
        gl_PointSize = 0.0;

    gl_Position = mvp * vert;


    //  logTemp = log(starTemp)*edlg10;
    float cc = cos(newRa)*sin(newDe);
    float z = zfactor(cc);
    logTempObs  = log( starTemp/z )*edlg10;
    logTempStar = log( starTemp )*edlg10;

    one_over_mu = pow(10.0,chi);

    // a freely falling observer outside the blackhole must transform
    // the static magnification to the freely falling magnification
    if (obsDist<1.0 && obsDist>=x0)
    {
        one_over_mu /= (gamma*gamma*(1.0-beta*cc)*(1.0-beta*cc));
    }
    else if (obsDist>1.0)
    {
        float a  = texture2D(ksiInsideArray, vec2(1.0/obsDist,phi*invMAX_PHI) ).r;
        float w  = sqrt(a*a-(1.0-obsDist)*obsDist*obsDist);
        float XX = obsDist*obsDist*sqrt(obsDist)+a*w;
        float YY = obsDist*sqrt(obsDist)*w-a*obsDist;
        float oldKsi = atan(YY,XX);

        float t2 = sqrt(obsDist-x0);
        float t3 = obsDist*t2;
        float t4 = sqrt(obsDist);
        float t5 = t4*obsDist;
        float t6 = cos(oldKsi);
        float t8 = t5*t6-obsDist;
        float t9 = t8*t8;
        float t10 = sin(oldKsi);
        float t11 = t10*t10;
        float t12 = 1.0/t11;
        float t15 = obsDist*obsDist;
        float t18 = sqrt(t9*t12-(1.0-obsDist)*t15);
        float t21 = t8/t10;
        float t28 = 2.0/t18*(-t21*t5-t9/t11/t10*t6);
        float t33 = sqrt(1.0-x0);
        float t35 = t8*t12;
        float t36 = obsDist*t33;
        float t43 = t15*t33*t2+t21*t18;
        float t48 = t3*t18-t21*t36;
        float t49 = t43*t43;
        float t50 = 1.0/t49;
        float t60 = t48*t48;
        float t64 = ((t3*t28/2.0+t4*t15*t33+t35*t36*t6)/t43-t48*t50*(-t5*t18-t35*t18*t6+t21*t28/2.0))/(1.0+t60*t50);

        one_over_mu *= sin(oldKsi)/sin(newKsi)*t64;
    }
    else
        one_over_mu = 0.0;
}

