/*
    dssbh_gui.cpp

  Copyright (c) 2009-2014  Thomas Mueller

    This file is part of DssBH.

    DssBH is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DssBH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DssBH.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QGridLayout>
#include <QGroupBox>
#include <QMenuBar>
#include <QMessageBox>

#include "dssbh_gui.h"

/*! Standard constructor.
*/
DBH_View :: DBH_View ( double vmax, QMainWindow* parent )
    : QMainWindow(parent)
{
    mVmagMax  = vmax;
    mCurrTime = 0.0;
    mTimeFactor = DEF_TIME_FACTOR;
    init();
}

/*! Standard destructor.
*/
DBH_View :: ~DBH_View() {
}

// ********************************** public methods *****************************

/*! Set maximum Vmag.
 *  \param  vmax : maximum visual apparent magnitude to be loaded
 *                 from the Hipparcos catalogue.
*/
void  
DBH_View :: setVmagMax ( double vmax )
{
    mVmagMax = vmax;
    if (mRender!=NULL)
        mRender->setVmagMax(vmax);
}


// *********************************** public slots ******************************

/*! Quit viewer.
*/
void  
DBH_View :: slot_reset ( )
{
    sli_currPosition->setValue(0);
    sli_currTime->setValue(0);
    mCurrTime = 0.0;

    double pos = spb_initPosition->value();
    led_currPosition->setText( QString::number(pos) );
}


/*! Quit viewer.
*/
void  
DBH_View :: slot_quit  ( )
{
    close();
}


/*! Show about.
 */
void  
DBH_View :: slot_about ( )
{
    QMessageBox::about( this, tr("Distort stellar sky by a Schwarzschild black hole") , tr("Copyright (c) 2009-2014, Thomas Mueller\nVisualization Research Center\nUniversity of Stuttgart, Germany\n\nEmail: Thomas.Mueller@vis.uni-stuttgart.de") );
}


/*! 
 */
void  
DBH_View :: slot_setInitValues ( )
{
    mRender->slot_setInitValues();
    mRender->slot_setInitPosition(spb_initPosition->value());
}


/*! 
 */
void  
DBH_View :: slot_change_mode ( )
{
    if (rab_obsStatic->isChecked())
    {
        sli_currPosition->setEnabled(false);
        led_currPosition->setEnabled(false);
        sli_currTime->setEnabled(false);
        led_currTime->setEnabled(false);
        pub_play->setEnabled(false);
        pub_reset->setEnabled(false);

        sli_currPosition->setValue(0);
        mCurrTime = 0.0;
    }
    else
    {
        sli_currPosition->setEnabled(true);
        led_currPosition->setEnabled(true);
        sli_currTime->setEnabled(true);
        led_currTime->setEnabled(true);
        pub_play->setEnabled(true);
        pub_reset->setEnabled(true);
    }
}


/*! 
 */
void  
DBH_View :: slot_update_crashTime ( )
{
    mCrashTime = mRender->calc_tauCrash();
    led_crashTime->setText( QString::number( mCrashTime ) );
}


/*! 
 */
void  
DBH_View :: slot_update_currTime  ( )
{
    mCurrTime =  (double)sli_currTime->value()/10000.5*mCrashTime;
    led_currTime->setText( QString::number( mCurrTime ) );

    double r0drs = mRender->getInitPosition();
    double t = mCurrTime*pow(r0drs,-1.5);

    double r;
    nrBrent( t_of_y, &t, 0.0, 1.0, 1e-7, r );
    led_currPosition->setText( QString::number( r*r0drs ) );
    mRender->setCurrPosition(r*r0drs);

    int slipos = 10000-(int)(r*10000.5);

    disconnect( sli_currPosition, 0, this, 0 );
    sli_currPosition->setValue(slipos);
    connect( sli_currPosition, SIGNAL( valueChanged(int) ), this, SLOT( slot_update_currPos() ) );
}


/*! 
 */
void  
DBH_View :: slot_update_currPos   ( )
{
    double r0drs = mRender->getInitPosition();
    double pos = (double)(10000.5-sli_currPosition->value())/10000.5;
    led_currPosition->setText( QString::number( pos*r0drs ) );
    mRender->setCurrPosition(pos*r0drs);

    double t = 0.0;
    double tau = t_of_y(pos,&t);
    led_currTime->setText( QString::number( tau*pow(r0drs,1.5) ) );

    int slipos = (int)(tau/M_PI_2*10000.5);
    disconnect( sli_currTime, 0, this, 0 );
    sli_currTime->setValue(slipos);
    connect( sli_currTime, SIGNAL( valueChanged(int) ), this, SLOT( slot_update_currTime() ) );
}


/*! 
 */
void  
DBH_View :: slot_setShaderParams  ( )
{
    double  gamma = spb_frag_gamma->value();
    double  f0    = spb_frag_f0->value();
    int     mps   = spb_vert_maxps->value();

    mRender->setShaderParams( gamma, f0, mps );
}


/*! 
 */
void  
DBH_View :: slot_startstop (bool anim)
{
    if (anim)
    {
        pub_play->setIcon( QIcon(":/pause.png") );
        mTimer->start();
    }
    else
    {
        pub_play->setIcon( QIcon(":/play.png") );
        mTimer->stop();
    }
}


/*! 
 */
void  
DBH_View :: slot_restart ( )
{
    slot_reset();
}


/*! 
*/
void    
DBH_View :: slot_animate ( )
{
    if (mCurrTime>=mCrashTime+0.01)
    {
        pub_play->setChecked(false);
        pub_play->setIcon( QIcon(":/play.png") );
        mTimer->stop();
        return;
    }
    mCurrTime += mTimeFactor;

    double r0drs = mRender->getInitPosition();
    double t = mCurrTime*pow(r0drs,-1.5);
    led_currTime->setText( QString::number( mCurrTime ) );

    double r;
    nrBrent( t_of_y, &t, 0.0, 1.0, 1e-7, r );
    led_currPosition->setText( QString::number( r*r0drs ) );
    mRender->setCurrPosition(r*r0drs);

    int slipos = 10000-(int)(r*10000.5);

    disconnect( sli_currPosition, 0, this, 0 );
    sli_currPosition->setValue(slipos);
    connect( sli_currPosition, SIGNAL( valueChanged(int) ), this, SLOT( slot_update_currPos() ) );

    slipos = (int)(t/M_PI_2*10000.5);
    disconnect( sli_currTime, 0, this, 0 );
    sli_currTime->setValue(slipos);
    connect( sli_currTime, SIGNAL( valueChanged(int) ), this, SLOT( slot_update_currTime() ) );
}


/*! 
*/
void    
DBH_View :: slot_changeFactor ( double val )
{
    mTimeFactor = val;
}


// ********************************* protected methods *****************************

/*!  Initialize all.
 */
void
DBH_View :: init()
{
    initElements();
    initGUI();

    initControl();
    initActions();
    initMenus();
}


/*!  Initialize GUI elements.
 */
void
DBH_View :: initElements()
{
    QGLFormat format;
    #if QT_VERSION >= 0x050000
        format.setVersion(3,3);
    #endif
    format.setProfile(QGLFormat::CoreProfile);
    format.setAlpha(true);

    mRender = new DBH_Render(format);
    mRender->setVmagMax( mVmagMax );


    mCrashTime = mRender->calc_tauCrash();
    mPosition  = mRender->getInitPosition();

    /* ---------------------------------
     *    observer elements
     * --------------------------------- */
    lab_initPosition = new QLabel("Init. pos. r_i/r_s");
    lab_initPosition->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_initPosition = new QDoubleSpinBox();
    spb_initPosition->setRange(1.01,1000.0);
    spb_initPosition->setSingleStep(0.01);
    spb_initPosition->setDecimals(2);
    spb_initPosition->setValue(mPosition);
    spb_initPosition->setMaximumHeight(DEF_WIDGET_HEIGHT);

    lab_obsMode = new QLabel("Mode");
    lab_obsMode->setMaximumHeight(DEF_WIDGET_HEIGHT);
    rab_obsStatic = new QRadioButton("quasi-static");
    rab_obsStatic->setMaximumHeight(DEF_WIDGET_HEIGHT);
    rab_obsStatic->setChecked(true);
    rab_obsFree   = new QRadioButton("freely falling");
    rab_obsFree->setMaximumHeight(DEF_WIDGET_HEIGHT);

    /* ---------------------------------
     *    animate elements
     * --------------------------------- */
    pub_play = new QPushButton( QIcon(":/play.png") ,"", this );
    pub_play->setMaximumHeight(DEF_WIDGET_HEIGHT+5);
    pub_play->setEnabled(false);
    pub_play->setCheckable(true);
    pub_reset = new QPushButton( QIcon(":/restart.png"), "", this );
    pub_reset->setMaximumHeight(DEF_WIDGET_HEIGHT+5);
    pub_reset->setEnabled(false);
    lab_anim_factor = new QLabel("factor");
    lab_anim_factor->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_anim_factor = new QDoubleSpinBox();
    spb_anim_factor->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_anim_factor->setDecimals(3);
    spb_anim_factor->setRange(1e-3,1e3);
    spb_anim_factor->setSingleStep(1e-3);
    spb_anim_factor->setValue(DEF_TIME_FACTOR);

    mTimer = new QTimer;
    mTimer->setInterval(30);
    connect( mTimer, SIGNAL(timeout()), this, SLOT( slot_animate() ) );

    /* ---------------------------------
     *    position/time elements
     * --------------------------------- */
    lab_crashTime     = new QLabel("Crash time");
    lab_crashTimeUnit = new QLabel("c" + QString(0x03c4) + "_crash/r_s");
    lab_crashTimeUnit->setMaximumHeight(DEF_WIDGET_HEIGHT);
    led_crashTime     = new QLineEdit( QString::number( mRender->calc_tauCrash() ));
    led_crashTime->setReadOnly(true);
    led_crashTime->setMaximumHeight(DEF_WIDGET_HEIGHT);

    lab_currPosition  = new QLabel("Curr. pos.");
    lab_currPosition->setMaximumHeight(DEF_WIDGET_HEIGHT);
    lab_currPositionUnit = new QLabel("r_obs/r_s");
    lab_currPositionUnit->setMaximumHeight(DEF_WIDGET_HEIGHT);
    sli_currPosition  = new QSlider( Qt::Horizontal );
    sli_currPosition->setMaximumHeight(DEF_WIDGET_HEIGHT);
    sli_currPosition->setRange(0,10000);
    sli_currPosition->setEnabled(false);
    led_currPosition  = new QLineEdit( QString::number(mPosition) );
    led_currPosition->setMaximumHeight(DEF_WIDGET_HEIGHT);
    led_currPosition->setReadOnly(true);
    led_currPosition->setEnabled(false);
    led_currPosition->setMaximumWidth(100);

    lab_currTime      = new QLabel("Curr. prop. time");
    lab_currTime->setMaximumHeight(DEF_WIDGET_HEIGHT);
    lab_currTimeUnit  = new QLabel("c" + QString(0x03c4));
    lab_currTimeUnit->setMaximumHeight(DEF_WIDGET_HEIGHT);
    sli_currTime      = new QSlider( Qt::Horizontal );
    sli_currTime->setMaximumHeight(DEF_WIDGET_HEIGHT);
    sli_currTime->setRange(0,10000);
    sli_currTime->setEnabled(false);
    led_currTime      = new QLineEdit( QString::number(0.0) );
    led_currTime->setMaximumHeight(DEF_WIDGET_HEIGHT);
    led_currTime->setReadOnly(true);
    led_currTime->setEnabled(false);
    led_currTime->setMaximumWidth(100);

    /* ---------------------------------
     *    shader elements
     * --------------------------------- */
    lab_frag_gamma = new QLabel("gamma");
    lab_frag_gamma->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_frag_gamma = new QDoubleSpinBox();
    spb_frag_gamma->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_frag_gamma->setRange(0.1,2.5);
    spb_frag_gamma->setSingleStep(0.01);
    spb_frag_gamma->setDecimals(2);
    spb_frag_gamma->setValue(DEF_FRAG_GAMMA);

    lab_frag_f0 = new QLabel("f0");
    lab_frag_f0->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_frag_f0 = new QDoubleSpinBox();
    spb_frag_f0->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_frag_f0->setRange(0.1,2.5);
    spb_frag_f0->setSingleStep(0.01);
    spb_frag_f0->setDecimals(2);
    spb_frag_f0->setValue(DEF_FRAG_F0);

    lab_vert_maxps = new QLabel("Max point size");
    lab_vert_maxps->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_vert_maxps = new QSpinBox();
    spb_vert_maxps->setMaximumHeight(DEF_WIDGET_HEIGHT);
    spb_vert_maxps->setRange(1,20);
    spb_vert_maxps->setValue(DEF_VERT_MPS);
}


/*!  Initialize GUI.
 */
void 
DBH_View :: initGUI()
{
    QGridLayout*  layout_complete = new QGridLayout();

    /* ---------------------------------
     *    layout observer
     * --------------------------------- */
    QGroupBox* grb_observer = new QGroupBox("Observer");
    QGridLayout* layout_observer = new QGridLayout();
    layout_observer -> addWidget( lab_initPosition, 0, 0 );
    layout_observer -> addWidget( spb_initPosition, 0, 1 );
    layout_observer -> addWidget( lab_obsMode,      1, 0 );
    layout_observer -> addWidget( rab_obsStatic,    1, 1 );
    layout_observer -> addWidget( rab_obsFree,      2, 1 );
    grb_observer->setLayout( layout_observer );


    /* ---------------------------------
     *    layout position/time
     * --------------------------------- */
    QGroupBox* grb_position = new QGroupBox("Position/Time");
    QGridLayout* layout_position = new QGridLayout();
    layout_position -> addWidget( lab_crashTime,        0, 0 );
    layout_position -> addWidget( lab_crashTimeUnit,    0, 1 );
    layout_position -> addWidget( led_crashTime,        0, 2 );
    layout_position -> addWidget( lab_currPosition,     1, 0 );
    layout_position -> addWidget( lab_currPositionUnit, 1, 1 );
    layout_position -> addWidget( sli_currPosition,     1, 2 );
    layout_position -> addWidget( led_currPosition,     1, 3 );
    layout_position -> addWidget( lab_currTime,         2, 0 );
    layout_position -> addWidget( lab_currTimeUnit,     2, 1 );
    layout_position -> addWidget( sli_currTime,         2, 2 );
    layout_position -> addWidget( led_currTime,         2, 3 );
    grb_position->setLayout( layout_position );


    /* ---------------------------------
     *    layout animate
     * --------------------------------- */
    QGroupBox* grb_animate = new QGroupBox("Animate");
    QGridLayout* layout_animate = new QGridLayout();
    layout_animate -> addWidget( pub_reset, 0, 0 );
    layout_animate -> addWidget( pub_play,  0, 1 );
    layout_animate -> addWidget( lab_anim_factor,  1, 0 );
    layout_animate -> addWidget( spb_anim_factor,  1, 1 );
    grb_animate->setLayout( layout_animate );


    /* ---------------------------------
     *    layout shader
     * --------------------------------- */
    QGroupBox* grb_shader = new QGroupBox("Shader");
    QGridLayout* layout_shader = new QGridLayout();
    layout_shader -> addWidget( lab_frag_gamma, 0, 0 );
    layout_shader -> addWidget( spb_frag_gamma, 0, 1 );
    layout_shader -> addWidget( lab_frag_f0,    1, 0 );
    layout_shader -> addWidget( spb_frag_f0,    1, 1 );
    layout_shader -> addWidget( lab_vert_maxps, 2, 0 );
    layout_shader -> addWidget( spb_vert_maxps, 2, 1 );
    grb_shader->setLayout( layout_shader );

    /* ---------------------------------
     *    layout bottom
     * --------------------------------- */
    QGridLayout* layout_bottom = new QGridLayout();
    layout_bottom -> addWidget( grb_observer, 0, 0 );
    layout_bottom -> addWidget( grb_position, 0, 1 );
    layout_bottom -> addWidget( grb_animate,  0, 2 );
    layout_bottom -> addWidget( grb_shader,   0, 3 );

    /* ---------------------------------
     *    layout_complete
     * --------------------------------- */
    layout_complete -> addWidget( mRender, 0, 0 );
    layout_complete -> addLayout( layout_bottom, 1, 0 );

    QWidget* widget = new QWidget();
    widget->setLayout( layout_complete );
    //  widget->setMinimumWidth(DEF_GUI_WIDTH );
    //  widget->setMaximumHeight( DEF_GUI_HEIGHT);
    setCentralWidget(widget);
}


/*!  Initialize connections.
 */
void
DBH_View :: initControl()
{ 
    connect( spb_initPosition, SIGNAL( valueChanged(double) ), mRender, SLOT( slot_setInitPosition(double) ) );
    connect( spb_initPosition, SIGNAL( valueChanged(double) ), this,   SLOT( slot_update_crashTime() ) );
    connect( spb_initPosition, SIGNAL( valueChanged(double) ), this,   SLOT( slot_reset() ) );

    connect( rab_obsStatic, SIGNAL( clicked(bool) ), this, SLOT( slot_change_mode() ) );
    connect( rab_obsFree,   SIGNAL( clicked(bool) ), this, SLOT( slot_change_mode() ) );
    connect( rab_obsStatic, SIGNAL( clicked(bool) ), mRender, SLOT( slot_setInitValues() ) );

    connect( sli_currPosition, SIGNAL( valueChanged(int) ), this, SLOT( slot_update_currPos() ) );
    connect( sli_currTime,     SIGNAL( valueChanged(int) ), this, SLOT( slot_update_currTime() ) );

    connect( pub_play,        SIGNAL( clicked(bool) ), this, SLOT( slot_startstop(bool) ) );
    connect( pub_reset,       SIGNAL( pressed() ),      this, SLOT( slot_restart() ) );
    connect( spb_anim_factor, SIGNAL( valueChanged(double) ), this, SLOT( slot_changeFactor(double) ) );

    connect( spb_frag_gamma, SIGNAL( valueChanged(double) ), this, SLOT( slot_setShaderParams() ) );
    connect( spb_frag_f0,    SIGNAL( valueChanged(double) ), this, SLOT( slot_setShaderParams() ) );
    connect( spb_vert_maxps, SIGNAL( valueChanged(int) ),    this, SLOT( slot_setShaderParams() ) );
}


/*!  Initialize actions.
 */
void
DBH_View :: initActions()
{
    /* ------------------
     *  View actions
     * ------------------ */
    /*
  mActionReloadShader = new QAction(QIcon(":/reload.png"),"Reload", this);
  mActionReloadShader->setShortcut(Qt::CTRL|Qt::Key_R);
  addAction( mActionReloadShader );
  connect  ( mActionReloadShader, SIGNAL(triggered()), mRender, SLOT( slot_reloadShaders() ) );
*/
    mActionReset = new QAction("Reset", this);
    mActionReset->setShortcut(Qt::Key_I);
    addAction( mActionReset );
    connect  ( mActionReset, SIGNAL(triggered()), this, SLOT( slot_setInitValues() ) );

    mActionQuit = new QAction(QIcon(":/exit.png"), "&Quit", this);
    mActionQuit->setShortcut(Qt::CTRL|Qt::Key_Q);
    addAction( mActionQuit );
    connect  ( mActionQuit, SIGNAL(triggered()), this, SLOT(slot_quit()) );



    /* ------------------
     *   Help actions
     * ------------------ */
    mActionAbout = new QAction(QIcon(":/about.png"),"&About", this );
    addAction( mActionAbout );
    connect ( mActionAbout, SIGNAL( triggered() ), this, SLOT( slot_about() ) );

}


/*!  Initialize menus.
 */
void
DBH_View :: initMenus()
{
    /* ---- View menu ---- */
    mViewMenu = menuBar() -> addMenu( "&View" );
    mViewMenu -> addAction( mActionReset );
    //mViewMenu -> addAction( mActionReloadShader );
    mViewMenu -> addSeparator();
    mViewMenu -> addAction( mActionQuit );

    /* ---- Help menu ---- */
    mHelpMenu = menuBar() -> addMenu( "&Help" );
    mHelpMenu -> addAction( mActionAbout );
}
