/*
    draw.cpp

  Copyright (c) 2009-2014  Thomas Mueller

    This file is part of DssBH.

    DssBH is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DssBH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DssBH.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QKeyEvent>
#include "dssbh_render.h"


int compare( const void *a, const void *b )
{
    float val_a = (*(skyStarCat_t*)a).plx;
    float val_b = (*(skyStarCat_t*)b).plx;

    // kleinste Parallaxe zuerst
    return ( (val_a-val_b)<0.0 );
}


/*! Standard constructor.
*/
DBH_Render :: DBH_Render(QGLFormat format,  QWidget* parent )
    : QGLWidget(format, parent)
{
    mButtonPressed = Qt::NoButton;
    setFocusPolicy( Qt::ClickFocus );
    mKeyPressed = Qt::Key_No;

    setMinimumSize( DEF_DRAW_WIDTH, DEF_DRAW_HEIGHT );
    setMaximumSize( DEF_DRAW_WIDTH, DEF_DRAW_HEIGHT );

    setMouseTracking(true);
    // setCursor( Qt::BlankCursor );

    mSigmaName           =  DEF_DATA_SIGMA;
    mPsiTempName         =  DEF_DATA_PSITEMP;
    mKsiArrayOutsideName = DEF_DATA_ARRAY_OUTSIDE;
    mKsiArrayInsideName  = DEF_DATA_ARRAY_INSIDE;

    mVertShaderName = DEF_VERT_SHADER_NAME;
    mFragShaderName = DEF_FRAG_SHADER_NAME;

    setInitValues();
    mDefGamma = (float)DEF_FRAG_GAMMA;
    mDefF0    = DEF_FRAG_F0;
    mDefMPS   = DEF_VERT_MPS;

    mVmagMax  = DEF_MAX_MAGNITUDE;
    va = vbo = mbo = tbo = 0;
}

/*! Standard destructor.
*/
DBH_Render :: ~DBH_Render ( )
{
}

// ************************************ public methods ********************************

/*! Set maximum Vmag.
 *  \param  vmax : maximum visual apparent magnitude to be loaded
 *                 from the Hipparcos catalogue.
*/
void    
DBH_Render :: setVmagMax ( double vmax )
{
    mVmagMax = vmax;
}


/*!  Read hipparcos data.
 *  \param filename : name of Hipparcos data file.
*/
bool
DBH_Render :: readHipBin ( std::string filename )
{
    long num_rows = 118218;
    mSkyCat  = new skyStarCat_t[ num_rows ];

    std::ifstream in(filename.c_str(),std::ios::binary);
    if (!in.is_open())
    {
        fprintf(stderr,"Can't open Hipparcos catalogue file '%s'.\n",filename.c_str());
        return false;
    }
    for(int i=0; i<num_rows; i++)
    {
        in.read((char*)&mSkyCat[i].hip,  sizeof(int));
        in.read((char*)&mSkyCat[i].ra,   sizeof(float));
        in.read((char*)&mSkyCat[i].dec,  sizeof(float));
        in.read((char*)&mSkyCat[i].plx,  sizeof(float));
        in.read((char*)&mSkyCat[i].Vmag, sizeof(float));
        in.read((char*)&mSkyCat[i].bmv,  sizeof(float));
    }

    // sort the catalogue with respect to trigonometric parallax;
    qsort( mSkyCat, num_rows, sizeof(skyStarCat_t), compare );

    mVerts        = new GLfloat[ num_rows*3 ];
    mMagnitudes   = new GLfloat[ num_rows ];
    mTemperatures = new GLfloat[ num_rows ];

    GLfloat *vptr = mVerts;
    GLfloat *mptr = mMagnitudes;
    GLfloat *tptr = mTemperatures;

    mNumStars = 0;
    for(int row=0; row<num_rows; row++)
    {
        if (mSkyCat[row].plx>0.0 && mSkyCat[row].Vmag<mVmagMax)
        {
            *(vptr++)  = GLfloat(mSkyCat[row].plx);
            *(vptr++)  = GLfloat(mSkyCat[row].ra)*DEG_TO_RAD;
            *(vptr++)  = GLfloat(mSkyCat[row].dec)*DEG_TO_RAD;
            *(mptr++)  = mSkyCat[row].Vmag + 5.0*log(1000.0/mSkyCat[row].plx)/log(10.0)-10.0;
            *(tptr++)  = calc_temp( mSkyCat[row].bmv );
            //          fprintf(stderr,"%6d %6.3f  %9.1f\n",row,mSkyCat[row].bmv,calc_temp( mSkyCat[row].bmv ));
            mNumStars++;
        }
    }
    fprintf(stderr,"#stars with Vmag < %3.2f : %d\n", mVmagMax, mNumStars);

    if (va>0) {
        glDeleteVertexArrays(1,&va);
    }

    glGenVertexArrays( 1, &va );
    glGenBuffers(1, &vbo);
    glGenBuffers(1, &mbo);
    glGenBuffers(1, &tbo);

    glBindVertexArray(va);

    glBindBuffer( GL_ARRAY_BUFFER, vbo );
    glEnableVertexAttribArray(0);
    glBufferData( GL_ARRAY_BUFFER, sizeof(float)*3*mNumStars, mVerts, GL_STATIC_DRAW);
    glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,0,NULL);

    glBindBuffer( GL_ARRAY_BUFFER, mbo );
    glEnableVertexAttribArray(1);
    glBufferData( GL_ARRAY_BUFFER, sizeof(float)*1*mNumStars, mMagnitudes, GL_STATIC_DRAW);
    glVertexAttribPointer(1,1,GL_FLOAT,GL_FALSE,0,NULL);

    glBindBuffer( GL_ARRAY_BUFFER, tbo );
    glEnableVertexAttribArray(2);
    glBufferData( GL_ARRAY_BUFFER, sizeof(float)*1*mNumStars, mTemperatures, GL_STATIC_DRAW);
    glVertexAttribPointer(2,1,GL_FLOAT,GL_FALSE,0,NULL);

    glBindVertexArray(0);
    return true;
}


/*! 
*/
double
DBH_Render :: calc_vel ( )
{
    if (rObsCurr>rObsInit || rObsCurr<=0.0 || x0>=1.0)
        return 0.0;
    return sqrt(fabs(rs/rObsCurr-x0)/(1.0-x0));
}


/*! 
*/
bool
DBH_Render :: calc_ksiCrit ( const double xObs, const double x0, double &ksiCrit )
{
    if (xObs>=x0 && x0<=1.0)
    {
        double p = 4.0/27.0;
        double z1 = xObs*xObs*sqrt(1.0-x0)*sqrt(xObs-x0);
        double z22 = fabs(p*(xObs*xObs*(xObs-1.0)+p));
        double z2 = sqrt(z22);
        double n  = xObs*xObs*(xObs-x0)+p;

        if (xObs>2.0/3.0)
            ksiCrit = acos((z1-z2)/n);
        else
            ksiCrit = acos((z1+z2)/n);
        return true;
    }
    return false;
}


/*! Get initial position.
 *  \return  initial position in units of r_s.
*/
double
DBH_Render :: getInitPosition   ( )
{
    return rObsInit/rs;
}


/*! Set current position.
 *  \param pos : current position in units of r_s.
*/
void
DBH_Render :: setCurrPosition   ( double pos )
{
    rObsCurr = pos*rs;
    updateGL();
}


/*! Crash time c*tau/rs = Pi/2 pow(rObsInit/rs,1.5)
*/
double
DBH_Render :: calc_tauCrash ( )
{
    return M_PI_2*pow(rObsInit/rs,1.5);
} 


void    
DBH_Render :: setShaderParams   ( double gamma, double f0, int mps )
{
    mDefGamma = (GLfloat)gamma;
    mDefF0    = (GLfloat)f0;
    mDefMPS = mps;
    updateGL();
}

// ************************************* public slots *********************************

/*! 
*/
void
DBH_Render :: slot_setInitValues  ( )
{
    setInitValues();
    updateGL();
}


/*! 
*/
void   
DBH_Render :: slot_reloadShaders   ( )
{
    std::cerr << "reload" << std::endl;
    initShaders();
    updateGL();
}


/*! 
*/
void   
DBH_Render :: slot_setInitPosition ( double pos )
{
    rObsInit = (float)pos*rs;
    x0 = rs/rObsInit;
    rObsCurr = rObsInit;
    updateGL();
}


/*! 
*/
void   
DBH_Render :: slot_resetFragParams ( )
{
    setInitFragParams();
    updateGL();
}


// *********************************** protected methods ******************************

/*! Initialize OpenGL rendering.
*/
void  
DBH_Render :: initializeGL ( )
{
    if (gl3wInit()) {
        fprintf(stderr,"OpenGL::Error: Failed to initialize gl3w.\n");
        exit(1);
    }
    fprintf(stderr,"Graphics board details:\n");
    fprintf(stderr,"\tVendor:         %s\n",glGetString(GL_VENDOR));
    fprintf(stderr,"\tGPU:            %s\n",glGetString(GL_RENDERER));
    fprintf(stderr,"\tOpenGL version: %s\n",glGetString(GL_VERSION));
    fprintf(stderr,"\tGLSL version:   %s\n\n",glGetString(GL_SHADING_LANGUAGE_VERSION));

    mMVPmatrix = glm::ortho(0.0f, (float)DEF_DRAW_WIDTH, 0.0f, (float)DEF_DRAW_HEIGHT);

    glViewport( 0, 0, DEF_DRAW_WIDTH, DEF_DRAW_HEIGHT );
    glClearColor( 0.0, 0.0, 0.0, 0.0 );

    glEnable( GL_DEPTH_TEST );
    glEnable( GL_BLEND );
    glBlendFunc( GL_ONE, GL_ONE );

    //glEnable( GL_POINT_SPRITE );
    glEnable(0x8861);

    initShaders();
    initTextures();

    readHipBin( std::string(DEF_HIPPARCOS_CATALOGUE_FILE_BIN) );
}


/*! Initialize shaders.
*/
void  
DBH_Render :: initShaders( )
{
    fprintf(stderr,"Init shaders...\n");
    if (!setShader( mVertShaderName, mFragShaderName, mProgram )) {
        fprintf(stderr,"cannot be initialized.\n");
    }
}


/*! Initialize textures.
*/
void  
DBH_Render :: initTextures ( )
{
    fprintf(stderr,"Init textures\n");
    int x,y,c;
    readBinaryWithHeader( mSigmaName, &mTexID_sigma, x, y, c );
    std::cerr << "Read " << mSigmaName << " ... " << x << " " << y << " " << c << std::endl;
    readBinaryWithHeader( mPsiTempName, &mTexID_psi, x, y, c );
    std::cerr << "Read " << mPsiTempName << " ... " << x << " " << y << " " << c << std::endl;
    readBinaryWithHeader( mKsiArrayOutsideName, &mTexID_outside, x, y, c );
    std::cerr << "Read " << mKsiArrayOutsideName << " ... " << x << " " << y << " " << c << std::endl;
    readBinaryWithHeader( mKsiArrayInsideName, &mTexID_inside, x, y, c );
    std::cerr << "Read " << mKsiArrayInsideName << " ... " << x << " " << y << " " << c << std::endl;
}


/*!  Main rendering method.
*/
void  
DBH_Render :: paintGL ( )
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glActiveTexture( GL_TEXTURE0 );
    glBindTexture( GL_TEXTURE_2D, mTexID_outside );

    glActiveTexture( GL_TEXTURE1 );
    glBindTexture( GL_TEXTURE_2D, mTexID_inside );

    glActiveTexture( GL_TEXTURE2 );
    glBindTexture( GL_TEXTURE_2D, mTexID_sigma );

    glActiveTexture( GL_TEXTURE3 );
    glBindTexture( GL_TEXTURE_2D, mTexID_psi );

    glUseProgram( mProgram );
    glUniformMatrix4fv( glGetUniformLocation( mProgram, "mvp"), 1, GL_FALSE, glm::value_ptr(mMVPmatrix) );

#if 0
    glUniform1i( glGetUniformLocation( mProgram,"ksiArray" ), 0 );
    glBegin( GL_QUADS );
    glTexCoord2f( 0, 0 );
    glVertex2f( 0, DEF_DRAW_HEIGHT );
    glTexCoord2f( 1, 0 );
    glVertex2f( DEF_DRAW_WIDTH, DEF_DRAW_HEIGHT );
    glTexCoord2f( 1, 1 );
    glVertex2f( DEF_DRAW_WIDTH, 0 );
    glTexCoord2f( 0, 1 );
    glVertex2f( 0, 0 );
    glEnd();
#endif

#if 1
    glUniform1i( glGetUniformLocation( mProgram,"ksiArray" ), 0 );
    glUniform1i( glGetUniformLocation( mProgram,"ksiInsideArray" ), 1 );
    glUniform1i( glGetUniformLocation( mProgram,"texSigma" ), 2 );
    glUniform1i( glGetUniformLocation( mProgram,"texPsiTemp" ), 3 );
    glUniform2f( glGetUniformLocation( mProgram,"wSize" ),  GLfloat(DEF_DRAW_WIDTH), GLfloat(DEF_DRAW_HEIGHT) );

    double diff = rs/rObsCurr - x0;
    double b = 1.0/rObsCurr*sqrt(1.0-x0);
    double a;
    if (fabs(diff)<1e-6) {
        a = 0.0;
    } else {
        a = 1.0/rObsCurr*sqrt(fabs(diff));
    }

    glUniform1f( glGetUniformLocation( mProgram,"obsDist" ), (GLfloat)rs/rObsCurr);
    glUniform1f( glGetUniformLocation( mProgram,"x0" ), x0);
    glUniform1f( glGetUniformLocation( mProgram,"a" ), GLfloat(a));
    glUniform1f( glGetUniformLocation( mProgram,"b" ), GLfloat(b));

    glUniformMatrix4fv( glGetUniformLocation( mProgram,"rotmat" ), 1, GL_FALSE, rotation_matrix );

    calc_ksiCrit(rs/rObsCurr, x0, mKsiCrit );
    glUniform1f( glGetUniformLocation( mProgram,"ksiCrit" ), (GLfloat)mKsiCrit );

    double beta = calc_vel();
    glUniform1f( glGetUniformLocation( mProgram,"beta"  ), GLfloat(beta) );
    glUniform1f( glGetUniformLocation( mProgram,"gamma" ), GLfloat(1.0/sqrt(1.0-beta*beta)) );

    glUniform1f( glGetUniformLocation( mProgram,"def_gamma" ), mDefGamma );
    glUniform1f( glGetUniformLocation( mProgram,"def_f0" ), mDefF0 );
    glUniform1f( glGetUniformLocation( mProgram,"def_max_pointsize" ), (GLfloat)mDefMPS );

    glUniform1i( glGetUniformLocation( mProgram,"order" ), 0 );
    drawPoints();

    glUniform1i( glGetUniformLocation( mProgram,"order" ), 1 );
    drawPoints();
#endif
    glUseProgram( 0 );


    glActiveTexture( GL_TEXTURE0 );
    glDisable( GL_TEXTURE_2D );
}


/*!  Main rendering method.
*/
void  
DBH_Render :: drawPoints ( )
{
    /*
   // glVertexPointer ( 3, GL_FLOAT, 0, mVerts );
    glVertexAttribPointer( TEMPERATURE_ARRAY, 1, GL_FLOAT, GL_FALSE, 0, mTemperatures );
    glVertexAttribPointer( MAGNITUDE_ARRAY, 1, GL_FLOAT, GL_FALSE, 0, mMagnitudes );

    //glEnableClientState ( GL_VERTEX_ARRAY );
    glEnableVertexAttribArray ( TEMPERATURE_ARRAY );
    glEnableVertexAttribArray ( MAGNITUDE_ARRAY );
    //  glPointSize(5.0);

    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
   // glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);
    glDrawArrays(GL_POINTS, 0, mNumStars);
    glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);

    glDisableVertexAttribArray ( TEMPERATURE_ARRAY );
    glDisableVertexAttribArray ( MAGNITUDE_ARRAY );
   // glDisableClientState ( GL_VERTEX_ARRAY );
   */

    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
    glBindVertexArray(va);
    glDrawArrays(GL_POINTS,0,mNumStars);
    glBindVertexArray(0);
    glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);
}


/*! Evaluate key press event.
*/
void 
DBH_Render :: keyPressEvent ( QKeyEvent* event )
{
    mKeyPressed = event->key();
    event->accept();
}


/*! Evaluate key release event.
*/
void 
DBH_Render :: keyReleaseEvent ( QKeyEvent* event )
{
    mKeyPressed = Qt::Key_No;
    event->accept();
}


/*! Evaluate mouse press event.
*/
void 
DBH_Render :: mousePressEvent ( QMouseEvent * event )
{
    mButtonPressed = event->button();
    last_pos = event->pos();
    oldAngle1 = rotationAngle1;
    oldAngle2 = rotationAngle2;
    updateGL();
}


/*! Evaluate mouse release event.
*/
void 
DBH_Render :: mouseReleaseEvent ( QMouseEvent * event )
{
    mButtonPressed = Qt::NoButton;
    event->accept();
    updateGL();
}


/*! Evaluate mouse move event.
*/
void 
DBH_Render :: mouseMoveEvent ( QMouseEvent * event )
{
    QPoint move = event->pos()-last_pos;

    if (mButtonPressed == Qt::LeftButton)
    {
        rotationAngle2 = oldAngle2-move.x()*0.01;
        recalcMatrix();
    }
    else if (mButtonPressed == Qt::MidButton)
    {
        rotationAngle1 = oldAngle1+move.y()*0.01;
        recalcMatrix();
    }
    updateGL();
}


/*! Set initial values
*/
void 
DBH_Render :: setInitValues ( )
{
    rs       = DEF_INIT_RS;
    rObsInit = DEF_INIT_POS*rs;
    rObsCurr = rObsInit;

    x0 = rs/rObsInit;
    mKsiCrit = 0.0;

    rotationAngle1 = 0.0;
    rotationAngle2 = 0.0;

    for(int i=0; i<16; i++)
        rotation_matrix[i] = 0.0;
    rotDir = 1.0;
    recalcMatrix();
}


/*! Recalculate rotation matrix.
*/
void    
DBH_Render :: recalcMatrix ( )
{
    double alpha = rotationAngle1;
    double beta  = rotationAngle2;

    rotation_matrix[0]  = cos(alpha)*cos(beta);
    rotation_matrix[1]  = cos(alpha)*sin(beta);
    rotation_matrix[2]  = sin(alpha);
    rotation_matrix[4]  = -sin(beta);
    rotation_matrix[5]  = cos(beta);
    rotation_matrix[6]  = 0.0;
    rotation_matrix[8]  = -sin(alpha)*cos(beta);
    rotation_matrix[9]  = -sin(alpha)*sin(beta);
    rotation_matrix[10] = cos(alpha);
}


/*! Calculate temperature from B-V value.
*/
float 
DBH_Render :: calc_temp ( const float bmv )
{
    double a = (bmv-C2)/C1;
    if (a<=3.961)
        return pow(10,a);

    double d = C4*C4-4.0*C3*(C5-bmv);
    if (d<0.0)
        return 0.0;

    double lgT = 0.5*(-C4-sqrt(d))/C3;
    return (float)pow(10.0,lgT);
} 


/*! Read binary data files with header.
*/
bool
DBH_Render :: readBinaryWithHeader ( std::string filename, GLuint* texID,  int &resX, int &resY, int &numChannels )
{
    std::ifstream in(filename.c_str(), std::ios::binary);
    if (!in.is_open())
    {
        fprintf(stderr,"Can't open file %s\n",filename.c_str());
        return false;
    }

    char buf[256];
    for(int i=0; i<4; i++)
    {
        buf[i] = in.peek();
        in.seekg(i+1);
    }

    if (strncmp(buf,"SIZE",4)==0)
    {
        in.read((char*)&resX,sizeof(int));  // resX
        in.read((char*)&resY,sizeof(int));  // resY
        in.read((char*)&numChannels,sizeof(int));  // numChannels
    }
    
    float* array = new float[resX*resY*numChannels];
    in.read( (char*) &array[0], sizeof(float)*resX*resY*numChannels );

    glGenTextures( 1, texID );
    glBindTexture( GL_TEXTURE_2D, *texID );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
    //glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    //glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

    if (numChannels==1) {
        glTexImage2D( GL_TEXTURE_2D, 0, GL_R32F, resX, resY, 0, GL_RED, GL_FLOAT, array );
    } else if (numChannels==2) {
        glTexImage2D ( GL_TEXTURE_2D, 0, GL_RG32F, resX, resY, 0, GL_RG, GL_FLOAT, array );
    } else if (numChannels==3) {
        glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB32F, resX, resY,  0, GL_RGB, GL_FLOAT, array );
    } else if (numChannels==4) {
        glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA32F, resX, resY, 0, GL_RGBA, GL_FLOAT, array );
    }

    in.close();

    delete [] array;
    return true;
}


/*! 
*/
void
DBH_Render :: setInitFragParams    ( )
{
    mDefGamma = (float)DEF_FRAG_GAMMA;
    mDefF0    = DEF_FRAG_F0;
    mDefMPS   = DEF_VERT_MPS;
    updateGL();
}

