/*
    dssbh_utils.cpp

  Copyright (c) 2009-2014  Thomas Mueller

    This file is part of DssBH.

    DssBH is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DssBH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DssBH.  If not, see <http://www.gnu.org/licenses/>.
*/

/*!   \file   dssbh_utils.cpp
      \brief  Shader tools.
*/

#ifndef M_PI_2
#define M_PI_2 1.5707963
#endif

#include "dssbh_utils.h"

#ifdef _WIN32
#include <io.h>
#include <sys\stat.h>
#endif

/*! Read shader from file.
 * \param shaderName : filename of shader
 * \return pointer to shader-text.
 */
char*  readShader  ( std::string shaderName )
{
    FILE *fp;
    char *content = NULL;

    int f,count;
    f = open(shaderName.c_str(), O_RDONLY);
    count = lseek(f, 0, SEEK_END);
    close(f);

    if (!shaderName.empty())
    {
        fp = fopen(shaderName.c_str(),"rt");

        if (fp != NULL)
        {
            if (count > 0) {
                content = (char *)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}

/*! Print log information about fragment- or vertex-shader./
 * \param shader : shader handle.
 */
bool   printShaderInfoLog   ( GLuint shader )
{
    int infoLogLen   = 0;
    int charsWritten = 0;
    GLchar *infoLog;

    int compileStatus = 0;
    glGetShaderiv(shader, GL_COMPILE_STATUS,  &compileStatus);
    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLogLen);

    if (!compileStatus && infoLogLen > 1)  {
        infoLog = new GLchar[infoLogLen];
        // error check for fail to allocate memory omitted
        glGetShaderInfoLog(shader, infoLogLen, &charsWritten, infoLog);
        printf("InfoLog : %s\n", infoLog);
        delete [] infoLog;
    }
    return (compileStatus == GL_TRUE);
}

/*! Print log information about shader-program./
 * \param shader : shader handle.
 */
bool   printProgramInfoLog ( GLuint progr  )
{
    int infoLogLen   = 0;
    int charsWritten = 0;
    GLchar *infoLog;

    int linkStatus = 0;
    glGetProgramiv(progr, GL_INFO_LOG_LENGTH, &infoLogLen);
    glGetProgramiv(progr, GL_LINK_STATUS, &linkStatus);

    if (!linkStatus && infoLogLen > 1)  {
        infoLog = new GLchar[infoLogLen];
        // error check for fail to allocate memory omitted
        glGetProgramInfoLog(progr, infoLogLen, &charsWritten, infoLog);
        printf("ProgramInfoLog : %s\n", infoLog);
        delete [] infoLog;
    }
    return (linkStatus == GL_TRUE);
}

/*! Set vertex and fragment shader and compile shader-program.
 * \param vertShaderName : filename of vertex shader.
 * \param fragShaderName : filename of fragment shader.
 * \param program : reference to shader-program handle.
 * \return true : shader-program was successfully compiled.
 * \return false : compile error.
 */
bool   setShader ( std::string vertShaderName, std::string fragShaderName, GLuint &program )
{
    char *vs = NULL, *fs = NULL;

    GLuint vertexShader   = glCreateShader( GL_VERTEX_SHADER   );
    GLuint fragmentShader = glCreateShader( GL_FRAGMENT_SHADER );
    vs = readShader(vertShaderName);
    fs = readShader(fragShaderName);

    const char * ff = fs;
    const char * vv = vs;
    glShaderSource( vertexShader, 1, &vv, NULL );
    glShaderSource( fragmentShader, 1, &ff, NULL );
    free(vs);free(fs);

    glCompileShader(vertexShader);
    glCompileShader(fragmentShader);

    printShaderInfoLog(vertexShader);
    printShaderInfoLog(fragmentShader);

    program = glCreateProgram();

    glAttachShader(program,fragmentShader);
    glAttachShader(program,vertexShader);

    glLinkProgram(program);
    printProgramInfoLog(program);

    return true;
}

/*! Relation between scaled time t and current position of freely falling observer.
 * \param y : scaled position r_obs/r0.
 * \param params : pointer to parameter t.
 */
double  t_of_y ( double y, void* params )
{
    if (y==0.0)
        return M_PI_2;

    double t = *(double*)params;
    double w = sqrt(1.0/y-1.0);
    return y*w + atan(w) - t;
}

/*! Brent's numerical root-finding algorithm.
 * \param func : function pointer.
 * \param params : parameter pointer.
 * \param x1 : left border.
 * \param x2 : right border.
 * \param tol : error tolerance.
 * \param x : reference to root.
 */
#define ITMAX 100
#define EPS 3.0e-14
int nrBrent(double (*func)(double, void*), void *params, double x1, double x2, double tol, double &x)
{
    int iter;
    double a=x1,b=x2,c=x2,d=0.0,e=0.0,min1,min2;
    double fa=(*func)(a,params),fb=(*func)(b,params),fc,p,q,r,s,tol1,xm;

    if ((fa > 0.0 && fb > 0.0) || (fa < 0.0 && fb < 0.0)) {
        x = 0.0;
        return 0;
        //nrerror("Root must be bracketed in nrbrent");
    }
    fc=fb;
    for (iter=1;iter<=ITMAX;iter++)
    {

        if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
            c=a;
            fc=fa;
            e=d=b-a;
        }
        if (fabs(fc) < fabs(fb)) {
            a=b;
            b=c;
            c=a;
            fa=fb;
            fb=fc;
            fc=fa;
        }
        tol1=2.0*EPS*fabs(b)+0.5*tol;
        xm=0.5*(c-b);
        if (fabs(xm) <= tol1 || fb == 0.0) {
            x = b;
            return 1;
        }
        if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
            s=fb/fa;
            if (a == c) {
                p=2.0*xm*s;
                q=1.0-s;
            } else {
                q=fa/fc;
                r=fb/fc;
                p=s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
                q=(q-1.0)*(r-1.0)*(s-1.0);
            }
            if (p > 0.0) q = -q;
            p=fabs(p);
            min1=3.0*xm*q-fabs(tol1*q);
            min2=fabs(e*q);
            if (2.0*p < (min1 < min2 ? min1 : min2)) {
                e=d;
                d=p/q;
            } else {
                d=xm;
                e=d;
            }
        } else {
            d=xm;
            e=d;
        }
        a=b;
        fa=fb;
        if (fabs(d) > tol1)
            b += d;
        else
            b += NR_SIGN(tol1,xm);
        fb=(*func)(b,params);
    }
    x = 0.0;
    return 0;
}
#undef ITMAX
#undef EPS
